package opti;

import java.util.Random;

import hilfsklassen.Parser;
import hilfsklassen.Dateioperationen;
import hilfsklassen.Aufruf;


import simulation.Antrieb;
import simulation.RungeKuttaIntegrator;

public class Optimierer
{
    Antrieb antrieb;
    RungeKuttaIntegrator rungekuttaintegrator;

    double tmax;
    double dt;
    int schritte;

    double[] yout;

    Random random = null;

    private double zufall(double von, double bis)
    {
        if(random==null)
            random = new Random(System.currentTimeMillis());
        double wert = 	random.nextDouble();

        return wert*(bis-von) + von;
    }

    public Optimierer()
    {
        antrieb = new Antrieb();
        rungekuttaintegrator = new RungeKuttaIntegrator();
        rungekuttaintegrator.add(antrieb);

        tmax = 10.0;
        dt   = 0.001;
        schritte = (int)(tmax/dt);

        yout = new double[schritte];
    }

    //Fehlerfunktion
    public double fehlerfunktion(double[] param)
    {
        antrieb.parP=param[0];
        antrieb.parI=param[1];
        antrieb.parD=param[2];
        
        antrieb.WSOLL = 20.0;

        double dt = 0.001;
        double  t = 0.0;
        double[] y;
        double[] yalt = {0.0,0.0,0.0};

        yout[0] = yalt[0]; //Anfangswert von omega

        for(int i=1;i<schritte;i++)
        {
            y = rungekuttaintegrator.zeitschritt(yalt,t,dt);

            yalt[0] = y[0];
            yalt[1] = y[1];
            yalt[2] = y[2];

            yout[i] = yalt[0]; //omega merken

            t+=dt;
        }

        double fehler = 0.0;

        for(int i=0;i<yout.length;i++)
        {
            double wert = antrieb.WSOLL - yout[i];
            if(wert<0.0)
                wert = -wert;
            fehler+=wert;
        }

        return fehler; 
    }

    public double[] optimieren(double[] start_param, double start_alfa,int anzahl_schritte)
    {
        int param_anzahl = start_param.length;

        //-2 -1 -0.5 0 0.5 1 2
        double[] faktoren = {-2.0,-1.0,-0.5,0.0,0.5,1.0,2.0};
        double[] merker = new double[7*param_anzahl];    
        double[] psatz_alt = new double[param_anzahl];
        double[] psatz_such = new double[param_anzahl];
        double[] alfa_alt = new double[param_anzahl];
        double[] alfa_neu = new double[param_anzahl];

        double[] psatz_neu = new double[param_anzahl];

        double[] fehler_richtung =  new double[param_anzahl];

        double fehler_akt;
        int    fehlerindex_akt;
        double param_akt;
        double fehler_best; //in einem Durchlauf gerade bester Fehler mit aktuell bestem Parametersatz

        for(int i=0;i<param_anzahl;i++)
        {
           psatz_alt[i] = start_param[i];
           alfa_alt[i] = start_alfa;
        }

        for(int i=0;i<anzahl_schritte;i++)
        {
            //1. Schritt
            //Fehler fr alle Varianten berechnen:
            for(int ii=0;ii<param_anzahl;ii++)
            {
                //psatz_such auf psatz_alt setzen:
                for(int pp=0;pp<param_anzahl;pp++)
                    psatz_such[pp] = psatz_alt[pp];

                for(int kk=0;kk<7;kk++)
                {
                    psatz_such[ii] = psatz_alt[ii] + faktoren[kk]*alfa_alt[ii];                                
                    merker[ii*7+kk] = fehlerfunktion(psatz_such);
                }
            }
            //2. Schritt
            //Bestimmung des neuen Parametersatzes und der neuen Alfa-Werte

//Unabhngigkeit der Richtungen nicht mehr annehmen!!!
//Nur eine Komponente pro Optimierungsschritt ndern.

            for(int ii=0;ii<param_anzahl;ii++)
            {
                //Index des kleinsten Fehlers bestimmen:
                fehler_akt = merker[ii*7+0];
                fehlerindex_akt = 0;
                for(int kk=1;kk<7;kk++)
                {
                    if(merker[ii*7+kk]<fehler_akt)
                    {
                        fehler_akt = merker[ii*7+kk];
                        fehlerindex_akt = kk;
                        fehler_richtung[ii] = fehler_akt;
                    }
                }
                //Neuen Parameter-Startpunkt festlegen fr die aktuelle Komponente
                psatz_neu[ii] = psatz_alt[ii] + faktoren[fehlerindex_akt]*alfa_alt[ii]; 
                //Neues Alfa bestimmen:
                alfa_neu[ii] = faktoren[fehlerindex_akt]*alfa_alt[ii];
                if(alfa_neu[ii]==0.0)
                {
                    //alfa_neu[ii] = 0.25*alfa_alt[ii];

                    for(int kk=0;kk<alfa_neu.length;kk++)
                        alfa_neu[kk] += zufall(-0.01,0.01); 
                }
//                if(alfa_neu[ii]<0.0)
//                    alfa_neu[ii] = -alfa_neu[ii];
            }

            //3. Schritt
            //Umkopieren fr neuen Durchlauf:
//NUR Komponente mit minimaler Fehlerrichtung nehmen:
            int index_komp=0;
            double fmin = fehler_richtung[0];
            for(int ii=1;ii<param_anzahl;ii++)
            {
                if(fehler_richtung[ii]<fmin)
                {
                    index_komp = ii;
                    fmin = fehler_richtung[ii];
                }
            }        

            psatz_alt[index_komp] = psatz_neu[index_komp]; 
            alfa_alt[index_komp]  = alfa_neu[index_komp];
/*
            for(int ii=0;ii<param_anzahl;ii++)
            {
                psatz_alt[ii] = psatz_neu[ii]; 
                alfa_alt[ii]  = alfa_neu[ii];
            }        
*/

            for(int ii=0;ii<param_anzahl;ii++)
            {
                psatz_neu[ii] = psatz_alt[ii]; 
                alfa_neu[ii]  = alfa_alt[ii];
            }        

            fehler_best = fehlerfunktion(psatz_neu);
            System.out.print("Nr."+i+" f="+fehler_best+"    ");

            for(int ii=0;ii<param_anzahl;ii++)
            {
                System.out.print(psatz_neu[ii]+" ");
            }
            System.out.println();
        }
        return psatz_neu;
    }

    private void speicherSimulation(double[] par,String dateiname)
    {
        antrieb.parP=par[0];
        antrieb.parI=par[1];
        antrieb.parD=par[2];
        
        antrieb.WSOLL = 20.0;

        double dt = 0.001;
        double  t = 0.0;
        double[] y;
        double[] yalt = {0.0,0.0,0.0};

        yout[0] = yalt[0]; //Anfangswert von omega

        StringBuffer sb = new StringBuffer();
        sb.append(t+" "+yalt[0]+"\n");

        for(int i=1;i<schritte;i++)
        {
            y = rungekuttaintegrator.zeitschritt(yalt,t,dt);

            yalt[0] = y[0];
            yalt[1] = y[1];
            yalt[2] = y[2];

            yout[i] = yalt[0]; //omega merken

            sb.append(t+" "+yalt[0]+"\n");

            t+=dt;
        }

        String data = sb.toString();
        data = Parser.ersetze(data,".",",");
        Dateioperationen.speichern("./"+dateiname,data);
    }

    public static void main(String[] args)
    {
        double start_alfa = 0.001;
        double[] start_param = {1.0,1.0,0.0};

        Optimierer optimierer = new Optimierer();

        double[] best_param = optimierer.optimieren(start_param,start_alfa,100);

        for(int i=0;i<start_param.length;i++)
        {
             System.out.println("Bester Parameter Nr."+i+": "+best_param[i]);
        }

        optimierer.speicherSimulation(start_param,"data_start.txt");
        optimierer.speicherSimulation(best_param,"data_best.txt");
        Aufruf.befehl("scilab -f datashow.sce");
    }
}


